import networkx as nx
import ipycytoscape
from IPython.core.display import display

node_centered = {
    'selector': 'node',
    'style': {
        'font-size': '10',
        'label': 'data(label)',
        'height': '60',
        'width': '60',
        'text-max-width': '60',
        'text-wrap': 'wrap',
        'text-valign': 'center',
        'background-color': 'data(color)',
        'background-opacity': 0.6,
        'border-width': 3,
        'border-color': '#D3D3D3'
    }
}
edge_directed_named = {
    'selector': 'edge',
    'style': {
        'font-size': '8',
        'label': 'data(label)',
        'line-color': '#D3D3D3',
        'text-rotation': 'autorotate',
        'target-arrow-shape': 'triangle',
        'target-arrow-color': '#D3D3D3',
        'curve-style': 'bezier',
        'text-background-color': "#FCFCFC",
        'text-background-opacity': 0.8,
        'text-background-shape': 'rectangle',
        'width': 'data(weight)'
    }
}
my_style = [node_centered, edge_directed_named]

try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse


def to_nextworkx(graph, colors, captions, mappings=None):
    networkx_graph = nx.MultiDiGraph()

    def add_node(node):
        label = ':' + ':'.join(node.labels)
        props = dict(node.items())
        color = ''
        if label in colors:
            color = colors[label]
        if label in captions and captions[label] in props:
            props['labels'] = label
            label = props[captions[label]]

        networkx_graph.add_node(node.id, label=label, color=color, properties=props, title=label, tooltip=str(props))

    def add_edge(edge):
        props = dict(edge.items())
        weight = props.get('weight', 1)
        networkx_graph.add_edge(edge.start_node.id, edge.end_node.id, weight=weight, label=edge.type, tooltip=str(props))

    for node in graph._nodes.values():
        add_node(node)

    for rel in graph._relationships.values():
        add_edge(rel)

    return networkx_graph


default_display_config = {'layout': 'dagre', 'padding': 0,
                          'nodeSpacing': 10, 'edgeLengthVal': 10,
                          'animate': True, 'randomize': True,
                          'maxSimulations': 1500}


def display_graph(networkx_graph, config):
    w = ipycytoscape.CytoscapeWidget()
    w.graph.add_graph_from_networkx(networkx_graph)
    w.set_style(my_style)
    w.set_layout(name=config['layout'] if 'layout' in config else default_display_config['layout'],
                 padding=config['padding'] if 'padding' in config else default_display_config['padding'],
                 nodeSpacing=config['nodeSpacing'] if 'nodeSpacing' in config else default_display_config[
                     'nodeSpacing'],
                 edgeLengthVal=config['edgeLengthVal'] if 'edgeLengthVal' in config else default_display_config[
                     'edgeLengthVal'],
                 animate=config['animate'] if 'animate' in config else default_display_config['animate'],
                 randomize=config['randomize'] if 'randomize' in config else default_display_config['randomize'],
                 maxSimulations=config['maxSimulations'] if 'maxSimulations' in config else default_display_config[
                     'maxSimulations'])
    w.set_tooltip_source('tooltip')
    display(w)


def run_query(driver, query, cfg, is_inline, params=None):
    # we return only the last one
    config = cfg['config']
    with driver.session(database=cfg['database']) as session:
        result = None
        for sub_query in query.split(';'):
            sub_query = sub_query.strip()
            if sub_query != '':
                result = session.run(sub_query, **params) if params else session.run(sub_query)

        graph = result.graph()
        if len(graph._nodes) > 0:
            nx_graph = to_nextworkx(graph, config['colors'], config['captions'])
            return nx_graph if is_inline else display_graph(nx_graph, config['layout'])
        else:
            return result.to_df()


def parse_arguments(args):
    parsed_uri = urlparse.urlparse(args.uri)

    uri = ''
    if parsed_uri.hostname and parsed_uri.scheme:
        uri = f'{parsed_uri.scheme}://{parsed_uri.hostname}'

    if parsed_uri.port:
        uri = f'{uri}:{parsed_uri.port}'

    auth = None
    if parsed_uri.username and parsed_uri.password:
        auth = (parsed_uri.username, parsed_uri.password)
    elif args.username and args.password:
        auth = (args.username, args.password)

    alias = args.alias if args.alias else 'default'

    db = parsed_uri.path[1:] if len(parsed_uri.path) > 0 != '' else None
    return {'uri': uri, 'auth': auth, 'alias': alias, 'database': db}





from argparse import Namespace

from .annotation_utils import get_decorators
from .utils import run_query, parse_arguments
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse
from neo4j import GraphDatabase
from IPython.core.magic import (
    Magics,
    cell_magic,
    line_magic,
    magics_class,
    needs_local_scope, no_var_expand
)
from IPython.core.magic_arguments import argument, magic_arguments
import json
import re
import ast

@magics_class
class Cy2Py(Magics):
    cypher_connections = {}
    config_cache = {}
    default_uri = 'bolt://localhost:7687'
    magic_args = None

    def __init__(self, shell):
        Magics.__init__(self, shell=shell)
        decorators = get_decorators(Cy2Py, 'argument')
        method_decorators = decorators['cypher']
        self.magic_args = dict([
            (decorator.args[0].s, (decorator.args[1].s, 'bool' if isinstance(decorator.keywords[0].value, ast.Str) else decorator.keywords[0].value.id))
            for decorator in method_decorators
        ])
        try:
            from google.colab import output
            output.enable_custom_widget_manager()
        except:
            None

    def __parse_args(self, line, local_ns):
        splitted = re.findall(r'[^"\s]\S*|".+?"', line)
        args = dict([(key[2:], None) for key in dict(self.magic_args.values()).keys()])
        inner_args = dict(self.magic_args.values())
        i = 0
        to_remove = []
        while i < len(splitted):
            arg = splitted[i]
            var_name = None

            arg_type = None
            arg_key = None
            if self.magic_args.__contains__(arg):
                arg_key, arg_type = self.magic_args[arg]
            elif inner_args.__contains__(arg):
                arg_type = inner_args[arg]
                arg_key = arg

            if not arg_key:
                i += 1
                continue

            arg_key = arg_key.replace('-', '')

            if 'bool' != arg_type:
                to_remove.append(splitted.pop(i))
                var_name = splitted[i]
                to_remove.append(splitted.pop(i))
            else:
                args[arg_key] = True
                to_remove.append(splitted.pop(i))
                continue

            var_name = var_name.strip()
            if var_name.startswith('"'):
                var_name = var_name[1:var_name.rfind('"')]
            elif var_name.startswith('\''):
                var_name = var_name[1:var_name.rfind('\'')]

            if var_name.startswith('$'):
                var_name = var_name[1:]
            elif var_name.startswith('{'):
                var_name = var_name[1:var_name.rfind('}')]

            if local_ns.__contains__(var_name):
                args[arg_key] = local_ns[var_name]
            elif local_ns.__contains__(var_name):
                args[arg_key] = local_ns[var_name]

        # we try to reconstruct the query
        if not args['query'] and len(to_remove) > 0:
            args['query'] = line
            for remove in to_remove:
                args['query'] = args['query'].replace(remove, '', 1)
            args['query'] = args['query'].strip()

        return Namespace(**args)

    def __manage_config(self, args, alias, parsed_args):
        if alias not in self.config_cache:
            self.config_cache[alias] = {'config': {
                'colors': {},
                'captions': {},
                'layout': {
                    'layout': 'dagre', 'padding': 0,
                    'nodeSpacing': 10, 'edgeLengthVal': 10,
                    'animate': True, 'randomize': True
                }
            }}

        config = self.config_cache[alias]['config']

        if args.colors:
            config['colors'] = args.colors if isinstance(args.colors, dict) else json.loads(json.loads(args.colors).replace('\'', '"'))

        if args.captions:
            config['captions'] = args.captions if isinstance(args.captions, dict) else json.loads(json.loads(args.captions).replace('\'', '"'))

        if args.layout:
            config['layout'] = args.layout if isinstance(args.layout, dict) else json.loads(json.loads(args.layout).replace('\'', '"'))

        self.config_cache[alias]['config'] = config

        if 'database' not in self.config_cache[alias]:
            self.config_cache[alias]['database'] = None
        if args.database:
            self.config_cache[alias]['database'] = args.database
        elif parsed_args['database'] and not self.config_cache[alias]['database']:
            self.config_cache[alias]['database'] = parsed_args['database']

        if parsed_args['uri'] != '':
            self.config_cache[alias]['uri'] = parsed_args['uri']

    @cell_magic
    @line_magic
    @needs_local_scope
    @magic_arguments()
    @argument("-a", "--alias", type=str, help="""
        The Neo4j connection configuration alias.
        You after you defined it the first time in combination with the other arguments
        you can use only it to connect to the Neo4j instance
        without repassing the arguments each time.
        If you don't specify it the first time you'll pass the config
        it'll became the default for each connection 
    """)
    @argument("-u", "--uri", type=str, help="""
        The Neo4j URI. You can use this kind of URI in order to define
        a specific database to query: neo4j://localhost:7687/my-db
    """)
    @argument("-us", "--username", type=str, help="The Neo4j User")
    @argument("-pw", "--password", type=str, help="The Neo4j Password")
    @argument("-db", "--database", type=str, help="The Neo4j Database, if not provided we use the default")
    @argument("-q", "--query", type=str, help="Pass the Cypher query as argument. Valid only when you use line magic.")
    @argument("-p", "--params", type=str, help="A map of Cypher query parameters")
    @argument("-c", "--close", type=str, help="Close a Driver connection by alias or URI")
    @argument("-co", "--colors", type=str, help="A map label/color")
    @argument("-ca", "--captions", type=str, help="A map label/caption")
    @argument("-la", "--layout", type=str, help="A map for layout configuration")
    @argument("-l", "--list", action='store_true', help="List active configurations")
    @no_var_expand
    def cypher(self, line, cell='', local_ns=None):
        # manage arguments
        is_inline = False
        args = self.__parse_args(line, local_ns)
        if cell == '' and args.query:
            cell = args.query
            is_inline = True
        if 'uri' not in args and 'list' not in args and 'close' not in args:
            return 'Pass --uri, --list or --close'

        if args.list:
            return self.cypher_connections.keys()
        elif args.close:
            driver = None
            if args.close in self.cypher_connections:
                driver = self.cypher_connections.pop(args.close)
            elif args.close in self.config_cache:
                driver = self.cypher_connections.pop(self.config_cache[args.close]['uri'], None)

            if driver:
                driver.close()
                return 'Driver closed successfully'
            else:
                return f'Connection not defined for {args.close}'

        parsed_args = parse_arguments(args)
        alias = parsed_args['alias']
        auth = parsed_args['auth']
        self.__manage_config(args, alias, parsed_args)
        uri = self.config_cache[alias]['uri']

        if uri not in self.cypher_connections:
            self.cypher_connections[uri] = GraphDatabase.driver(uri, auth=auth)

        return run_query(self.cypher_connections[uri], cell, self.config_cache[alias], is_inline, args.params)


def load_ipython_extension(ip):
    ip.register_magics(Cy2Py)



import ast
import inspect


def get_decorators(cls, *args):
    target = cls
    decorators = {}

    def visit_function_def(node):
        decorators[node.name] = []
        for n in node.decorator_list:
            name = ''
            if isinstance(n, ast.Call):
                name = n.func.attr if isinstance(n.func, ast.Attribute) else n.func.id
            else:
                name = n.attr if isinstance(n, ast.Attribute) else n.id

            if len(args) == 0 or args.__contains__(name):
                decorators[node.name].append(n)

    node_iter = ast.NodeVisitor()
    node_iter.visit_FunctionDef = visit_function_def
    node_iter.visit(ast.parse(inspect.getsource(target)))
    return decorators






from unittest import TestCase

from IPython.testing.globalipapp import start_ipython
from testcontainers.neo4j import Neo4jContainer

from src.cy2py.cy2 import Cy2Py


class CypherMagicTest(TestCase):
    neo4j_container = None
    neo4j_version = '4.4.10-enterprise'
    neo4j_url = None
    _ip = None

    @classmethod
    def setUpClass(cls):
        cls.neo4j_container = (Neo4jContainer('neo4j:' + cls.neo4j_version)
                               .with_env("NEO4J_ACCEPT_LICENSE_AGREEMENT", "yes")
                               .with_env("NEO4J_dbms_security_auth__enabled", "false"))
        cls.neo4j_container.start()
        cls.neo4j_url = cls.neo4j_container.get_connection_url()
        with cls.neo4j_container.get_driver().session() as session:
            session.write_transaction(
                lambda tx: tx.run("UNWIND range(1, 10) AS id CREATE (c:Customer{id: id, name: 'Person ' + id })"))

        cls._ip = start_ipython()
        cls._ip.register_magics(Cy2Py)

    @classmethod
    def tearDownClass(cls):
        try:
            cls.neo4j_container.stop()
        except:
            pass

    def test_line_magic(self):
        res = self._ip.run_cell(raw_cell=f'''
            %cypher -l
        ''')
        assert res.success
        self._ip.run_cell(raw_cell=f'''
            neo4j_url = '{self.neo4j_url}'
        ''')
        res = self._ip.run_cell(raw_cell='''
            df = %cypher -u $neo4j_url MATCH (c:Customer) RETURN c.companyName as name
            print([[df['name'][ind] for ind in df.index]])
        ''')

        assert res.success
        self._ip.run_cell(raw_cell=f'''
            query = 'MATCH (c:Customer) RETURN c.companyName as name'
        ''')
        res = self._ip.run_cell(raw_cell='''
            df = %cypher -q "$query"
            print([[df['name'][ind] for ind in df.index]])
        ''')
        assert res.success
        res = self._ip.run_cell(raw_cell='''
            graph_query = 'MATCH (c:Customer) RETURN c'
            nx_graph = %cypher -q $graph_query
            print(nx_graph)
        ''')
        assert res.success
        res = self._ip.run_cell(raw_cell='''
            %cypher -c $neo4j_url
        ''')
        assert res.success

    def test_cell_magic(self):
        self._ip.run_cell(raw_cell='''
            colors = {"Product": "blue"}
        ''')
        self._ip.run_cell(raw_cell=f'''
            neo4j_url = '{self.neo4j_url}'
        ''')
        res = self._ip.run_cell(raw_cell='''
            %%cypher -u $neo4j_url -co "$colors"
            MATCH (c:Customer)
            RETURN c
        ''')
        assert res.success
        # do the second call by leveraging the cache
        res = self._ip.run_cell(raw_cell='''
            %%cypher
            MATCH (c:Customer)
            RETURN c
        ''')
        assert res.success

    def test_cell_magic_non_default_db(self):
        self._ip.run_cell(raw_cell='''
            colors = {"Product": "blue"}
        ''')
        self._ip.run_cell(raw_cell=f'''
            neo4j_url = '{self.neo4j_url}/system'
        ''')
        res = self._ip.run_cell(raw_cell='''
            %%cypher -u $neo4j_url -co "$colors"
            SHOW DATABASES
        ''')
        assert res.success
        # do the second call by leveraging the cache
        res = self._ip.run_cell(raw_cell='''
            %%cypher
            SHOW DATABASES
        ''')
        assert res.success



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

install_requires = [
    "ipython>=1.0",
    "neo4j>=4.4.5",
    "ipython-genutils>=0.1.0",
    "ipycytoscape>=1.3.3",
    "networkx"
]

setup(
    name="cy2py",
    version="1.1.3",
    author="Andrea Santurbano",
    author_email="santand@gmail.com",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/conker84/cy2py",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    install_requires=install_requires
)
